@@ -1,5 +1,5 @@
 /*
- * Copyright 1995-2025 The OpenSSL Project Authors. All Rights Reserved.
+ * Copyright 1995-2024 The OpenSSL Project Authors. All Rights Reserved.
  * Copyright 2005 Nokia. All rights reserved.
  *
  * Licensed under the Apache License 2.0 (the "License").  You may not use
@@ -109,7 +109,6 @@ SSL_SESSION *SSL_SESSION_new(void)
     if (ss == NULL)
         return NULL;
 
-    ss->ext.max_fragment_len_mode = TLSEXT_max_fragment_length_UNSPECIFIED;
     ss->verify_result = 1;      /* avoid 0 (= X509_V_OK) just in case */
    /* 5 minute timeout by default */
     ss->timeout = ossl_seconds2time(60 * 5 + 4);
@@ -139,12 +138,7 @@ static SSL_SESSION *ssl_session_dup_intern(const SSL_SESSION *src, int ticket)
     dest = OPENSSL_malloc(sizeof(*dest));
     if (dest == NULL)
         return NULL;
-
-    /*
-     * src is logically read-only but the prev/next pointers are not, they are
-     * part of the session cache and can be modified concurrently.
-     */
-    memcpy(dest, src, offsetof(SSL_SESSION, prev));
+    memcpy(dest, src, sizeof(*dest));
 
     /*
      * Set the various pointers to NULL so that we can call SSL_SESSION_free in
@@ -521,7 +515,7 @@ SSL_SESSION *lookup_sess_in_cache(SSL_CONNECTION *s,
     if (ret == NULL && s->session_ctx->get_session_cb != NULL) {
         int copy = 1;
 
-        ret = s->session_ctx->get_session_cb(SSL_CONNECTION_GET_USER_SSL(s),
+        ret = s->session_ctx->get_session_cb(SSL_CONNECTION_GET_SSL(s),
                                              sess_id, sess_id_len, &copy);
 
         if (ret != NULL) {
@@ -590,8 +584,6 @@ int ssl_get_prev_session(SSL_CONNECTION *s, CLIENTHELLO_MSG *hello)
     SSL_TICKET_STATUS r;
 
     if (SSL_CONNECTION_IS_TLS13(s)) {
-        SSL_SESSION_free(s->session);
-        s->session = NULL;
         /*
          * By default we will send a new ticket. This can be overridden in the
          * ticket processing.
@@ -604,7 +596,6 @@ int ssl_get_prev_session(SSL_CONNECTION *s, CLIENTHELLO_MSG *hello)
                                         hello->pre_proc_exts, NULL, 0))
             return -1;
 
-        /* If we resumed, s->session will now be set */
         ret = s->session;
     } else {
         /* sets s->ext.ticket_expected */
@@ -846,7 +837,7 @@ void SSL_SESSION_free(SSL_SESSION *ss)
     if (ss == NULL)
         return;
     CRYPTO_DOWN_REF(&ss->references, &i);
-    REF_PRINT_COUNT("SSL_SESSION", i, ss);
+    REF_PRINT_COUNT("SSL_SESSION", ss);
     if (i > 0)
         return;
     REF_ASSERT_ISNT(i < 0);
@@ -880,7 +871,7 @@ int SSL_SESSION_up_ref(SSL_SESSION *ss)
     if (CRYPTO_UP_REF(&ss->references, &i) <= 0)
         return 0;
 
-    REF_PRINT_COUNT("SSL_SESSION", i, ss);
+    REF_PRINT_COUNT("SSL_SESSION", ss);
     REF_ASSERT_ISNT(i < 2);
     return ((i > 1) ? 1 : 0);
 }
@@ -951,14 +942,19 @@ long SSL_SESSION_get_timeout(const SSL_SESSION *s)
 
 long SSL_SESSION_get_time(const SSL_SESSION *s)
 {
-    if (s == NULL)
-        return 0;
-    return (long)ossl_time_to_time_t(s->time);
+    return (long) SSL_SESSION_get_time_ex(s);
 }
 
-long SSL_SESSION_set_time(SSL_SESSION *s, long t)
+time_t SSL_SESSION_get_time_ex(const SSL_SESSION *s)
 {
-    OSSL_TIME new_time = ossl_time_from_time_t((time_t)t);
+    if (s == NULL)
+        return 0;
+    return ossl_time_to_time_t(s->time);
+}
+
+time_t SSL_SESSION_set_time_ex(SSL_SESSION *s, time_t t)
+{
+    OSSL_TIME new_time = ossl_time_from_time_t(t);
 
     if (s == NULL)
         return 0;
@@ -976,6 +972,11 @@ long SSL_SESSION_set_time(SSL_SESSION *s, long t)
     return t;
 }
 
+long SSL_SESSION_set_time(SSL_SESSION *s, long t)
+{
+    return (long) SSL_SESSION_set_time_ex(s, (time_t) t);
+}
+
 int SSL_SESSION_get_protocol_version(const SSL_SESSION *s)
 {
     return s->ssl_version;
